/*****************************************************************************
* wanpipe_voicetime.c	WANPIPE VoiceTime - USB Voice Sync Timer Interface
*
*
* Authors:	Alex Feldman
*
* Copyright:	(c) 2008 Sangoma Technologies Inc.
*
* ============================================================================
* June 10, 2008	Alex Feldman	Initial version.
*****************************************************************************/

/***************************************************************************
****		I N C L U D E  		F I L E S			****
***************************************************************************/


# include "wanpipe_includes.h"

#if defined(__LINUX__)
#include <linux/tty.h>
#include <linux/usb.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
	# include <linux/usb/input.h>
#else
	/*input.h is really not used in 1.0.5....uses a function to decode the ID values to the registered names*/
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
	# include <linux/hid.h>
#else
	# include <linux/hiddev.h>
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
	# include <linux/usb/serial.h>
#else
	# define MAX_NUM_PORTS 8  //This is normally defined in /linux/usb/serial.h
#endif
#include <linux/input.h>

#endif

# include "wanpipe_defines.h"
# include "wanpipe_version.h"
# include "wanpipe_debug.h"
# include "wanpipe_common.h"

#if defined (__WINDOWS__)
# include <wanpipe\csu_dsu.h>
#else
# include "zapcompat.h" /* Map of Zaptel -> DAHDI definitions */
#endif

#define LINUX26
#if 0
# define WVT_DEBUG		/* Remove it before release */
#endif
/***************************************************************************
****		   	  M A C R O S / D E F I N E S			****
***************************************************************************/
#define WVT_VENDOR_ID		0x10C4	
#define WVT_PRODUCT_ID		0x8460	
#define WVT_MOD_MAJOR_VER	1
#define WVT_MOD_MINOR_VER	1
#define WVT_MOD_NAME		"wanpipe_voicetime"
#define WVT_MOD_DESC		"WANPIPE VoiceTime (USB)"
#define WVT_VERSION		"1.0.9"

#define WVT_NAME		"wvt"
#define WVT_NAME_LEN		20

#define WVT_CLOCK_SRC		"wanpipe_voicetime"

#define WVT_STATUS_READY	1
#define WVT_STATUS_DOWN		2

#define WVT_MAX_RX_LEN		1				// number of bytes per ms
#define WVT_RXBYTES_SEC		(WVT_MAX_RX_LEN * 1000)		// number of bytes per second		

#define WVT_STATS_RANGE		10000
#define WVT_STATS_LOW		(95 * WVT_STATS_RANGE) 
#define WVT_TICKS_LOW		(HZ-35)
#define WVT_STATS_RX_SKIP	1000

#define WVT_STATS_SAMPLE	10				// every 30 sec
#define WVT_STATS_MAX_ERRS	5

#define WVT_TRANSFER_BULK	1
#define WVT_TRANSFER_INTR	2

/***************************************************************************
****		     S T R U C T U R E S / T Y P E D E F S		****
***************************************************************************/
typedef struct wud_softc_
{
	char				name[WVT_NAME_LEN];
	char 				phys[64];		//hid
	int				transfer_type;

	struct usb_interface		*usb_intf;
	struct usb_device		*usb_dev;
	struct input_dev 		*idev;
	int				usb_class;

	unsigned int			status;
        struct urb			*rx_urb;
	unsigned char			*rxdata;
	dma_addr_t			rxdata_dma;		//hid
	int				rxcount;
	int 				rx_bytes;
	wan_time_t 			last_sec;
	wan_suseconds_t			last_usec;
	int				err_cnt;
	int				stat_samples;

	struct zt_span			span;
	struct zt_chan*			chan;
	struct zt_chan*			chan_ptrs[1];
	struct zt_chan			_chan;

	WAN_LIST_ENTRY(wud_softc_)	next;

} wud_softc_t;

struct wud_desc {
        char	*name;
        int	adptr_type;
	int	transfer_type;
};

/***************************************************************************
****			F U N C T I O N  P R O T O T Y P E S		****
***************************************************************************/
static int wanpipe_voicetime_init(void* arg);
//int wanpipe_voicetime_shutdown(void *arg);
//int wanpipe_voicetime_ready_unload(void *arg);
static int wanpipe_voicetime_exit (void *arg);

static int wud_probe(struct usb_interface*, const struct usb_device_id*);
static void wud_disconnect(struct usb_interface*);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10) 
	static int wud_suspend (struct usb_interface*, pm_message_t);
#else
	static int wud_suspend (struct usb_interface*, u32 msg);
#endif
static int wud_resume (struct usb_interface*);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
	static int wud_prereset (struct usb_interface*);
	static int wud_postreset (struct usb_interface*);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
	static void wud_prereset (struct usb_interface*);
	static void wud_postreset (struct usb_interface*);
#endif


static int wud_setup(struct usb_interface*);
static int wud_remove(struct usb_interface*);
static int wud_initialize(wud_softc_t*);
static int wud_start_transfer(wud_softc_t*, struct urb*);
static unsigned long wud_get_score(unsigned int rx_bytes, unsigned int exp_bytes);
static int wud_setup_cp2102_urbs(wud_softc_t *sc);
#if 0
static int wud_setup_hid(wud_softc_t *sc);
#endif

/***************************************************************************
****			   G L O B A L  V A R I A B L E S			****
***************************************************************************/
static unsigned long	score_best = 0, score_worst = 100 * WVT_STATS_RANGE;

static struct wud_desc wud_dinfo = { WVT_MOD_DESC, 0, WVT_TRANSFER_BULK };
//static struct wud_desc wud_dinfo2 = { WVT_MOD_DESC, 0, WVT_TRANSFER_INTR };

WAN_LIST_HEAD(, wud_softc_) wud_softc_head = 
			WAN_LIST_HEAD_INITIALIZER(&wud_softc_head);

static struct usb_device_id wud_dev_ids[] = {

	{ 
		match_flags: (USB_DEVICE_ID_MATCH_VENDOR|USB_DEVICE_ID_MATCH_DEVICE),
		bInterfaceClass: USB_CLASS_AUDIO,
		bInterfaceSubClass: 1,
		idVendor:	WVT_VENDOR_ID,
		idProduct: 	WVT_PRODUCT_ID,
		driver_info:	(unsigned long)&wud_dinfo,
	},
	{ }     /* Terminating Entry */
};

static struct usb_driver wud_driver =
{
	name: 		WVT_MOD_NAME,
	probe: 		wud_probe,
	disconnect:	wud_disconnect,
	suspend:	wud_suspend,
	resume:		wud_resume,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
	pre_reset:	wud_prereset,
	post_reset:	wud_postreset,
#endif
	id_table:	wud_dev_ids,
};


#if defined(__LINUX__) || defined(__FreeBSD__)
WAN_MODULE_DEFINE(
		wanpipe_voicetime, "wanpipe_voicetime",
		"Alex Feldman <alex@sangoma.com>",
		WVT_MOD_DESC,
		"GPL",
		wanpipe_voicetime_init, wanpipe_voicetime_exit,
		NULL);
WAN_MODULE_VERSION(wanpipe_voicetime, WVT_MOD_MAJOR_VER);
#endif

/***************************************************************************
****			 F U N C T I O N  D E F I N I T I O N		****
***************************************************************************/


int wanpipe_voicetime_init(void* arg)
{
	int		ret;

	DEBUG_EVENT("%s: Loading %s Driver - v%s\n",
			WVT_MOD_NAME, WVT_MOD_DESC, WVT_VERSION);
	WAN_LIST_INIT(&wud_softc_head);

	/* USB-FXO */
	ret = usb_register(&wud_driver);
	if (ret){
		DEBUG_EVENT("%s: Failed to register %s device!\n",
				WVT_MOD_NAME, WVT_MOD_DESC);
		return ret;
	}

	return 0;
}

#if 0
int wanpipe_voicetime_shutdown(void *arg)
{
	DEBUG_EVENT("%s: Shutting down %s \n",
			WVT_MOD_NAME, WVT_MOD_DESC);
	return 0;
}

int wanpipe_voicetime_ready_unload(void *arg)
{
	DEBUG_EVENT("%s: Is %s ready to unload \n",
			WVT_MOD_NAME, WVT_MOD_DESC);
	return 0;
}
#endif

/*============================================================================
 * Module deinit point.
 * o release all remaining system resources
 */
int wanpipe_voicetime_exit (void *arg)
{
	wud_softc_t	*sc = NULL;

	DEBUG_EVENT("%s: Wanpipe VoiceTime result: Best: %ld.%ld%% Worst: %ld.%ld%%\n",
			WVT_MOD_NAME,
			score_best / WVT_STATS_RANGE, score_best % WVT_STATS_RANGE, 
			score_worst / WVT_STATS_RANGE, score_worst % WVT_STATS_RANGE); 

	/* Mark all devices down */
	WAN_LIST_FOREACH(sc, &wud_softc_head, next){
		wan_set_bit(WVT_STATUS_DOWN, &sc->status);
	}
	
	usb_deregister(&wud_driver);

	DEBUG_EVENT("%s: %s Driver Unloaded\n",
			WVT_MOD_NAME, WVT_MOD_DESC);
	return 0;
}


/****************************************************************************
****************************************************************************/
static int
wud_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
	struct usb_device	*dev = interface_to_usbdev(intf);
	struct wud_desc 	*dinfo = (struct wud_desc *)id->driver_info;
	wud_softc_t		*sc = NULL;

	DEBUG_EVENT("%s: Probing %s device on %d ...\n",
				WVT_MOD_NAME, 
				dinfo->name, dev->devnum);

	if ((sc= usb_get_intfdata(intf)) != NULL){
		DEBUG_EVENT("%s: Force to add usb device!\n",
				WVT_MOD_NAME);
	}

	if (wud_setup(intf)){
		DEBUG_EVENT("%s: Failed to setup Wanpipe VoiceTime device!\n",
				WVT_MOD_NAME);
		return -ENODEV;
	}

	return 0;
}

static void wud_disconnect(struct usb_interface *intf)
{
	struct usb_device	*dev = interface_to_usbdev(intf);

	DEBUG_EVENT("%s: Disconnecting %s device from %d ...\n",
				WVT_MOD_NAME, 
				WVT_MOD_DESC, dev->devnum);
	wud_remove(intf);
	return;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
static int wud_suspend (struct usb_interface *intf, pm_message_t message)
{
	struct usb_device	*dev = interface_to_usbdev(intf);

	DEBUG_EVENT("%s: WARNING: Suspend USB device on %d \n",
				WVT_MOD_NAME, dev->devnum);
	return 0;
}
#else
static int wud_suspend (struct usb_interface *intf, u32 msg)
{
	struct usb_device	*dev = interface_to_usbdev(intf);

	DEBUG_EVENT("%s: WARNING: Suspend USB device on %d \n",
				WVT_MOD_NAME, dev->devnum);
	return 0;
}
#endif

static int wud_resume (struct usb_interface *intf)
{
	struct usb_device	*dev = interface_to_usbdev(intf);

	DEBUG_EVENT("%s: WARNING: Resume USB device on %d \n",
				WVT_MOD_NAME, dev->devnum);
	return 0;
}

#if defined(LINUX26)  &&LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23))
static int wud_prereset (struct usb_interface *intf)
#else
static void wud_prereset (struct usb_interface *intf)
#endif
{
	struct usb_device	*dev = interface_to_usbdev(intf);

	DEBUG_EVENT("%s: WARNING: Pre-reset USB device on %d \n",
				WVT_MOD_NAME, dev->devnum);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23))
	return 0;
#else
	return;
#endif
}
#endif

#if defined(LINUX26)  &&LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23))
static int wud_postreset (struct usb_interface *intf)
#else
static void wud_postreset (struct usb_interface *intf)
#endif
{
	struct usb_device	*dev = interface_to_usbdev(intf);

	DEBUG_EVENT("%s: WARNING: Post-Reset USB device on %d \n",
				WVT_MOD_NAME, dev->devnum);
	
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23))
	return 0;
#else
	return;
#endif
}
#endif

/****************************************************************************
****************************************************************************/

static unsigned long wud_get_score(unsigned int rx_bytes, unsigned int exp_bytes)
{
	unsigned long	percent;

	if (rx_bytes == 0) {
		return 0;	// should never happened
	}
	percent = (100 * WVT_STATS_RANGE * abs(rx_bytes - exp_bytes)) / rx_bytes;
	return (100 * WVT_STATS_RANGE - abs(percent));
}

/* Calculate VoiceTime statistics */
static int wud_rx_stats(wud_softc_t *sc)
{
	int		bytes;
	wan_time_t 	sec;
	wan_suseconds_t	usec;
	unsigned long	score;

	wan_getcurrenttime(&sec, &usec);
	bytes =  (sec - sc->last_sec) * 1000;
	bytes += (usec - sc->last_usec) / 125; 

	sc->stat_samples++;
	score = wud_get_score(sc->rx_bytes, bytes);
	if (score < WVT_STATS_LOW){
		sc->err_cnt++;
		if (sc->stat_samples > WVT_STATS_SAMPLE){
			if (sc->err_cnt > WVT_STATS_MAX_ERRS){
				DEBUG_EVENT("%s: Wanpipe VoiceTime overruns counter: %d\n",
					sc->name, sc->err_cnt);
			}
			sc->err_cnt = 0;
			sc->stat_samples = 0;
		}
	}
#if defined(WVT_DEBUG)
	DEBUG_EVENT("%s: Wanpipe VoiceTime stats: %ld.%ld%%\n",
			sc->name, score / WVT_STATS_RANGE, score % WVT_STATS_RANGE);
#endif
	if (score > score_best) score_best = score;
	if (score < score_worst) score_worst = score;

	sc->last_sec  = sec;
	sc->last_usec = usec;
	sc->rx_bytes = 0;
	return 0;	
}

#if defined(LINUX26)  && (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))
void wud_read_complete(struct urb *q, struct pt_regs *regs)
#else
void wud_read_complete(struct urb *q)
#endif
{
	wud_softc_t	*sc = (wud_softc_t*)q->context;

	WAN_ASSERT_VOID(sc == NULL);
	sc->rxcount++;
	sc->rx_bytes += q->actual_length;

	if (wan_test_bit(WVT_STATUS_DOWN, &sc->status)){
		goto read_complete_done;
	}
	if (!wan_test_bit(WVT_STATUS_READY, &sc->status)){
		DEBUG_EVENT("%s: ERROR: RX USB core is not ready!\n",
					sc->name);
		goto read_complete_done;
	}
	if (q->actual_length != WVT_MAX_RX_LEN){
		DEBUG_TEST("%s: WARNING: Invalid Rx length (%d:%d)\n",
					sc->name, 
					q->actual_length,
					WVT_MAX_RX_LEN);
	}

	if (sc->rxcount < WVT_STATS_RX_SKIP){
		// init statistics 
		wan_getcurrenttime(&sc->last_sec, &sc->last_usec);
		sc->rx_bytes = 0;
		sc->err_cnt  = 0;
	}
	if (sc->rx_bytes > WVT_RXBYTES_SEC){
		if (wud_rx_stats(sc)){
			DEBUG_EVENT("%s: Wanpipe VoiceTime overruns counter: %d\n",
					sc->name, sc->err_cnt);
			sc->err_cnt = 0;
		}
	}

	zt_receive(&sc->span);
	zt_transmit(&sc->span);

	if (wud_start_transfer(sc, q)){
		DEBUG_RX(
		"%s: Failed to re-start RX transfer request!\n",
					sc->name);
		goto read_complete_done;
	}

read_complete_done:
	return;
}

static int wud_start_transfer(wud_softc_t *sc, struct urb *q)
{
	int	ret;

	if (wan_test_bit(WVT_STATUS_DOWN, &sc->status)){
		return -EINVAL; 
	}

#ifdef LINUX26
	ret = usb_submit_urb(q, GFP_KERNEL); 
#else
	ret = usb_submit_urb(q); 
#endif
	if (ret){
		DEBUG_TEST(
		"%s: Failed to issue transfer request (%d)!\n",
					sc->name, ret);
		return -EINVAL;
	}
	return 0;
}


#define usb_endpoint_dir_in(epd)					\
		(((epd)->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN)
#define usb_endpoint_dir_out(epd)					\
		(((epd)->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_OUT)
#define usb_endpoint_xfer_bulk(epd)					\
		(((epd)->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) ==	\
		USB_ENDPOINT_XFER_BULK)
#define usb_endpoint_is_bulk_in(epd)					\
		(usb_endpoint_xfer_bulk((epd)) && usb_endpoint_dir_in(epd))
#define usb_endpoint_is_bulk_out(epd)					\
		(usb_endpoint_xfer_bulk((epd)) && usb_endpoint_dir_out(epd))

static int wud_setup_cp2102_urbs(wud_softc_t *sc)
{
	struct usb_host_interface	*iface_desc;
	struct usb_endpoint_descriptor	*endpoint;
	struct usb_endpoint_descriptor	*bulk_in_endpoint[MAX_NUM_PORTS];
	int				num_bulk_in = 0;
	int				i = 0;

	WAN_ASSERT(sc == NULL);
	WAN_ASSERT(sc->usb_intf == NULL);

	/* descriptor matches, let's find the endpoints needed */
	/* check out the endpoints */
	iface_desc = sc->usb_intf->cur_altsetting;
	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		endpoint = &iface_desc->endpoint[i].desc;

		if (usb_endpoint_is_bulk_in(endpoint)) {
			/* we found a bulk in endpoint */
			DEBUG_TEST("%s: Found bulk in on endpoint %d\n", 
						sc->name, i);
			bulk_in_endpoint[num_bulk_in] = endpoint;
			++num_bulk_in;
		}
	}

	DEBUG_TEST("%s: Bulk in: %d MaxLen: %d\n",
				sc->name, num_bulk_in, WVT_MAX_RX_LEN);
	/* set up the endpoint information */
	for (i = 0; i < num_bulk_in; ++i) {
		endpoint = bulk_in_endpoint[i];
		sc->rxdata = wan_malloc(endpoint->wMaxPacketSize + 1);
		if (sc->rxdata == NULL){
			DEBUG_EVENT("%s: Failed to allocate rx buffer!\n", sc->name);
			return -ENOMEM;
		}
		usb_init_urb(sc->rx_urb);
		usb_fill_bulk_urb (
				sc->rx_urb,
				sc->usb_dev,
				usb_rcvbulkpipe (sc->usb_dev,
						   endpoint->bEndpointAddress),
				sc->rxdata, 
				WVT_MAX_RX_LEN,
				wud_read_complete,
				sc);	//p
		DEBUG_TEST("%s: Bulk Int ENAddress %X\n",
					sc->name,
					endpoint->bEndpointAddress);
	}
	
	return 0;
}

#define REQTYPE_HOST_TO_DEVICE	0x41
#define REQTYPE_DEVICE_TO_HOST	0xc1

/* Config SET requests. To GET, add 1 to the request number */
#define CP2101_UART 		0x00	/* Enable / Disable */
#define CP2101_BAUDRATE		0x01	/* (BAUD_RATE_GEN_FREQ / baudrate) */
#define CP2101_BITS		0x03	/* 0x(0)(databits)(parity)(stopbits) */
#define CP2101_BREAK		0x05	/* On / Off */
#define CP2101_CONTROL		0x07	/* Flow control line states */
#define CP2101_MODEMCTL		0x13	/* Modem controls */
#define CP2101_CONFIG_6		0x19	/* 6 bytes of config data ??? */

/* CP2101_UART */
#define UART_ENABLE		0x0001
#define UART_DISABLE		0x0000

/* CP2101_BAUDRATE */
#define BAUD_RATE_GEN_FREQ	0x384000

/* CP2101_BITS */
#define BITS_DATA_MASK		0X0f00
#define BITS_DATA_5		0X0500
#define BITS_DATA_6		0X0600
#define BITS_DATA_7		0X0700
#define BITS_DATA_8		0X0800
#define BITS_DATA_9		0X0900

#define BITS_PARITY_MASK	0x00f0
#define BITS_PARITY_NONE	0x0000
#define BITS_PARITY_ODD		0x0010
#define BITS_PARITY_EVEN	0x0020
#define BITS_PARITY_MARK	0x0030
#define BITS_PARITY_SPACE	0x0040

#define BITS_STOP_MASK		0x000f
#define BITS_STOP_1		0x0000
#define BITS_STOP_1_5		0x0001
#define BITS_STOP_2		0x0002


static int 
wud_set_config(wud_softc_t *sc, u8 request, unsigned int *data, int size)
{
//struct sdla_usb_pvt *p;
	__le32 *buf;
	int result, i, length;

	WAN_ASSERT(sc->usb_dev == NULL);
	/* Number of integers required to contain the array */
	length = (((size - 1) | 3) + 1)/4;

	buf = kmalloc(length * sizeof(__le32), GFP_KERNEL);
	if (!buf) {
		DEBUG_EVENT("%s: ERROR: Failed to allocate memory (%ld bytes)!\n",
					sc->name, (unsigned long)(length*sizeof(__le32)));
		return -ENOMEM;
	}

	/* Array of integers into bytes */
	for (i = 0; i < length; i++)
		buf[i] = cpu_to_le32(data[i]);

	if (size > 2) {
		result = usb_control_msg (sc->usb_dev,
				usb_sndctrlpipe(sc->usb_dev, 0),
				request, REQTYPE_HOST_TO_DEVICE, 0x0000,
				0, buf, size, 300);
	} else {
		result = usb_control_msg (sc->usb_dev,
				usb_sndctrlpipe(sc->usb_dev, 0),
				request, REQTYPE_HOST_TO_DEVICE, data[0],
				0, NULL, 0, 300);
	}

	kfree(buf);

	if ((size > 2 && result != size) || result < 0) {
		DEBUG_EVENT("%s: ERROR: Failed to send request (%X:%d:%d)!\n",
					sc->name, request, size, result);
		return -EPROTO;
	}

	/* Single data value */
	result = usb_control_msg(sc->usb_dev,
				usb_sndctrlpipe(sc->usb_dev, 0),
				request, REQTYPE_HOST_TO_DEVICE, data[0],
				0, NULL, 0, 300);
	return 0;
}


/*
 * cp2101_get_config
 * Reads from the CP2101 configuration registers
 * 'size' is specified in bytes.
 * 'data' is a pointer to a pre-allocated array of integers large
 * enough to hold 'size' bytes (with 4 bytes to each integer)
 */
static int 
wud_get_config(wud_softc_t *sc, u8 request, unsigned int *data, int size)
{
//struct sdla_usb_pvt *p;
	__le32 *buf;
	int result, i, length;

	WAN_ASSERT(sc->usb_dev == NULL);
	/* Number of integers required to contain the array */
	length = (((size - 1) | 3) + 1)/4;

	buf = kcalloc(length, sizeof(__le32), GFP_KERNEL);
	if (!buf) {
		DEBUG_EVENT("%s: ERROR: Failed allocate memory (%ld bytes)!\n",
					sc->name, (unsigned long)(length*sizeof(__le32)));
		return -ENOMEM;
	}

	/* For get requests, the request number must be incremented */
	request++;

	/* Issue the request, attempting to read 'size' bytes */
	result = usb_control_msg (	sc->usb_dev,
					usb_rcvctrlpipe (sc->usb_dev, 0),
					request, REQTYPE_DEVICE_TO_HOST, 0x0000,
					0, buf, size, 300);

	/* Convert data into an array of integers */
	for (i=0; i<length; i++)
		data[i] = le32_to_cpu(buf[i]);

	kfree(buf);

	if (result != size) {
		DEBUG_EVENT("%s: ERROR: Unable to send config request (%X:%d:%d)!\n",
					sc->name, request, size, result);
		return -EPROTO;
	}

	return 0;
}

/*
 * cp2101_set_config_single
 * Convenience function for calling cp2101_set_config on single data values
 * without requiring an integer pointer
 */
static inline int 
wud_set_config_single(wud_softc_t *sc, u8 request, unsigned int data)
{
	return wud_set_config(sc, request, &data, 2);
}

static int wud_setup_cp2102(wud_softc_t *sc)
{
	unsigned int	baud, bits;

	if (wud_set_config_single(sc, CP2101_UART, UART_ENABLE)) {
		DEBUG_EVENT("%s: ERROR: Failed to enable UART!\n",
					sc->name);
		return -EINVAL;
	}

//	baud = 512000;	//9600;
//	baud = 500000;	//9600;
	baud = 921600;	//115200;
	if (wud_set_config_single(sc, CP2101_BAUDRATE, (BAUD_RATE_GEN_FREQ / baud))){
		DEBUG_EVENT("%s: ERROR: Failed to set new baud rate (%d bps)!\n",
					sc->name, baud);
		return -EINVAL;
	}

	wud_get_config(sc, CP2101_BITS, &bits, 2);
	bits &= ~BITS_DATA_MASK;
	bits |= BITS_DATA_8;
	if (wud_set_config(sc, CP2101_BITS, &bits, 2)){
		DEBUG_EVENT("%s: ERROR: Failed set new data bits (%X)!\n",
					sc->name, bits);
		return -EINVAL;
	}

	//wan_spin_lock_irq_init(&sc->ilock, WVT_MOD_NAME);

	if (wud_setup_cp2102_urbs(sc)) {
		DEBUG_EVENT("%s: ERROR: Failed to prepare the urbs for transfer!\n",
					sc->name);
		return -EINVAL;
	}
	sc->usb_class = USB_CLASS_VENDOR_SPEC;
	return 0;
}

#if 0
static int wud_setup_hid(wud_softc_t *sc)
{
	struct input_dev *input_dev;
	struct usb_endpoint_descriptor *endpoint;
	struct usb_host_interface *interface;
	int pipe, maxp, status;

	WAN_ASSERT(sc == NULL);
	WAN_ASSERT(sc->usb_intf == NULL);

	interface = sc->usb_intf->cur_altsetting;
	if (interface->desc.bNumEndpoints != 1){
		return -ENODEV;
	}

	endpoint = &interface->endpoint[0].desc;
	if (!usb_endpoint_is_int_in(endpoint)){
		return -ENODEV;
	}

	sc->rxdata = usb_buffer_alloc(sc->usb_dev, 8, GFP_ATOMIC, &sc->rxdata_dma);
	if (!sc->rxdata){
		return -ENOMEM;
	}

	usb_make_path(sc->usb_dev, sc->phys, sizeof(sc->phys));
	strlcat(sc->phys, "/input0", sizeof(sc->phys));

	input_dev = input_allocate_device();
	if (!input_dev){
		usb_buffer_free(sc->usb_dev, 8, sc->rxdata, sc->rxdata_dma);
		DEBUG_EVENT("%s: Failed allocate input dev!\n",
				WVT_MOD_NAME);
		return -ENOMEM;
	}
	sc->idev = input_dev;

	input_dev->name = sc->name;
	input_dev->phys = sc->phys;
	usb_to_input_id(sc->usb_dev, &input_dev->id);
	input_dev->dev.parent = &sc->usb_intf->dev;

	pipe = usb_rcvintpipe(sc->usb_dev, endpoint->bEndpointAddress);

	usb_fill_int_urb(sc->rx_urb, sc->usb_dev, pipe, sc->rxdata, WVT_MAX_RX_LEN,
			 wud_read_complete, sc, endpoint->bInterval);
 	sc->rx_urb->transfer_dma = sc->rxdata_dma;
	sc->rx_urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

	if (input_register_device(sc->idev)){
		DEBUG_EVENT("%s: Failed allocate input dev!\n",
					WVT_MOD_NAME);
		input_free_device(input_dev);
		usb_buffer_free(sc->usb_dev, 8, sc->rxdata, sc->rxdata_dma);
		return -ENODEV;
	}
	sc->usb_class = USB_CLASS_HID;
	return 0;
}
#endif

static int wud_setup(struct usb_interface *intf)
{
	struct usb_device	*dev = interface_to_usbdev(intf);
	wud_softc_t		*sc = NULL;
	int			err;

	DEBUG_TEST("%s: Creating private data...\n", WVT_MOD_NAME);
	sc = wan_malloc(sizeof(wud_softc_t));
	if (!sc) {
		DEBUG_EVENT("%s: Failed allocate private data!\n",
				WVT_MOD_NAME);
		return -ENODEV;
	}
	memset(sc, 0, sizeof(wud_softc_t));
	sprintf(sc->name, "%s0", WVT_NAME);
	sc->usb_intf	= intf;
	sc->usb_dev	= dev;

	sc->rx_urb = usb_alloc_urb(0, GFP_KERNEL);
	if (!sc->rx_urb){
		DEBUG_EVENT("%s: Failed allocate irq!\n",
				WVT_MOD_NAME);
		wan_free(sc);
		return -ENOMEM;
	}

	err = wud_setup_cp2102(sc);
	//err = wud_setup_hid(sc);

	if (err){
		usb_free_urb(sc->rx_urb);
		wan_free(sc);
		return -ENODEV;
	}

	usb_set_intfdata(sc->usb_intf, sc);
	WAN_LIST_INSERT_HEAD(&wud_softc_head, sc, next);
	wan_set_bit(WVT_STATUS_READY, &sc->status);

	if (wud_initialize(sc)){
		DEBUG_EVENT("%s: Failed to register into Zaptel driver!\n",
					sc->name);
		goto wud_setup_error;
	}

	if (wud_start_transfer(sc, sc->rx_urb)) {
		DEBUG_EVENT("%s: Failed to start RX transfer!\n",
					sc->name);
		zt_unregister(&sc->span);
		goto wud_setup_error;
	}
	return 0;

wud_setup_error:

	WAN_LIST_REMOVE(sc, next);
	wan_clear_bit(WVT_STATUS_READY, &sc->status);
	if (sc->usb_class == USB_CLASS_HID){
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10) 
		input_free_device(sc->idev);
#else
		input_unregister_device(sc->idev);
#endif
		usb_buffer_free(sc->usb_dev, 8, sc->rxdata, sc->rxdata_dma);
	}else{
		if (sc->rxdata) wan_free(sc->rxdata);
	}
	usb_set_intfdata(sc->usb_intf, NULL);
	if (sc) wan_free(sc);
	return -ENODEV;
}

static int wud_remove(struct usb_interface *intf)
{
	wud_softc_t	*sc;

	DEBUG_TEST("Releasing private data...\n");
	if ((sc = usb_get_intfdata(intf)) != NULL){

		wan_clear_bit(WVT_STATUS_READY, &sc->status);

		zt_unregister(&sc->span);

		WAN_LIST_REMOVE(sc, next);
		usb_unlink_urb(sc->rx_urb);
		usb_kill_urb(sc->rx_urb);

		if (sc->usb_class == USB_CLASS_HID){
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10) 
		input_free_device(sc->idev);
#else
		input_unregister_device(sc->idev);
#endif
			usb_buffer_free(sc->usb_dev, 8, sc->rxdata, sc->rxdata_dma);
		}else{
			if (sc->rxdata) wan_free(sc->rxdata);
		}
		usb_set_intfdata(sc->usb_intf, NULL);
		if (sc) wan_free(sc);
	}else{

		DEBUG_EVENT("%s: Internal Error: [%s:%d]\n",
				WVT_MOD_NAME, __FUNCTION__,__LINE__);
	}
	return 0;
}

static int wud_initialize(wud_softc_t *sc)
{
	/* Zapata stuff */
	sc->chan = &sc->_chan;
	sc->chan_ptrs[0] = &sc->_chan; 

	sprintf(sc->span.name, "WANVTIME/1");
	snprintf(sc->span.desc, sizeof(sc->span.desc) - 1, "%s (source: " WVT_CLOCK_SRC ") %d", sc->span.name, 1);
	sprintf(sc->chan->name, "WANVTIME/%d/%d", 1, 0);
	sc->chan->chanpos = 1;
#ifdef DAHDI_ISSUES    
	sc->span.chans = sc->chan_ptrs;
#else
	sc->span.chans = sc->chan;
#endif
	sc->span.channels = 0;		/* no channels on our span */
	sc->span.deflaw = ZT_LAW_MULAW;
	init_waitqueue_head(&sc->span.maintq);
	sc->span.pvt = sc;
	sc->chan->pvt = sc;
	if (zt_register(&sc->span, 0)) {
		return -1;
	}
	return 0;
}
